import 'dart:math';

import 'package:bonfire/joystick/joystick_controller.dart';
import 'package:bonfire/util/mixins/movement.dart';

/// Mixin responsible for adding movements through joystick events
mixin MovementByJoystick on Movement {
  static const REDUCTION_SPEED_DIAGONAL = 0.7;

  /// flag to set if you only want the 8 directions movement. Set to false to have full 360 movement
  bool dPadAngles = true;

  /// the angle the player should move in 360 mode
  double movementRadAngle = 0;

  @override
  void update(double dt) {
    if (this is JoystickListener) {
      bool joystickContainThisComponent =
          gameRef.joystick?.containObserver(this as JoystickListener) ?? false;

      var newAngle = innerCurrentDirectionalAngle;
      if (dPadAngles || newAngle == 0.0) {
        newAngle = _getAngleByDirectional();
      }
      if (innerCurrentDirectional != JoystickMoveDirectional.IDLE &&
          newAngle != 0.0) {
        movementRadAngle = newAngle;
      }

      if (dPadAngles) {
        if (innerCurrentDirectional != null && joystickContainThisComponent) {
          final diagonalSpeed = this.speed * REDUCTION_SPEED_DIAGONAL;
          _moveDirectional(innerCurrentDirectional!, speed, diagonalSpeed);
        }
      } else {
        if (innerCurrentDirectional != null && joystickContainThisComponent) {
          if (innerCurrentDirectional != JoystickMoveDirectional.IDLE) {
            moveFromAngle(speed, movementRadAngle);
          }
          // movement was done on the above line, this is only for the animation
          // which is why we use zero speed as we don't want to translate position twice
          _moveDirectional(innerCurrentDirectional!, 0, 0);
        }
      }
    }

    super.update(dt);
  }

  void _moveDirectional(
    JoystickMoveDirectional direction,
    double speed,
    double diagonalSpeed,
  ) {
    switch (direction) {
      case JoystickMoveDirectional.MOVE_UP:
        moveUp(speed);
        break;
      case JoystickMoveDirectional.MOVE_UP_LEFT:
        moveUpLeft(diagonalSpeed, diagonalSpeed);
        break;
      case JoystickMoveDirectional.MOVE_UP_RIGHT:
        moveUpRight(diagonalSpeed, diagonalSpeed);
        break;
      case JoystickMoveDirectional.MOVE_RIGHT:
        moveRight(speed);
        break;
      case JoystickMoveDirectional.MOVE_DOWN:
        moveDown(speed);
        break;
      case JoystickMoveDirectional.MOVE_DOWN_RIGHT:
        moveDownRight(diagonalSpeed, diagonalSpeed);
        break;
      case JoystickMoveDirectional.MOVE_DOWN_LEFT:
        moveDownLeft(diagonalSpeed, diagonalSpeed);
        break;
      case JoystickMoveDirectional.MOVE_LEFT:
        moveLeft(speed);
        break;
      case JoystickMoveDirectional.IDLE:
        if (!isIdle) {
          idle();
        }
        break;
    }
  }

  /// get currentDirectional from `JoystickListener`
  JoystickMoveDirectional? get innerCurrentDirectional {
    if (this is JoystickListener) {
      return (this as JoystickListener).currentDirectional;
    } else {
      print(
          '(MovementByJoystick) ERROR: $this need use JoystickListener mixin');
      return null;
    }
  }

  /// get currentDirectional from `JoystickListener`
  double get innerCurrentDirectionalAngle {
    if (this is JoystickListener) {
      return (this as JoystickListener).currentDirectionalAngle;
    } else {
      return 0;
    }
  }

  double _getAngleByDirectional() {
    switch (innerCurrentDirectional) {
      case JoystickMoveDirectional.MOVE_LEFT:
        return 180 / (180 / pi);
      case JoystickMoveDirectional.MOVE_RIGHT:
        // we can't use 0 here because then no movement happens
        // we're just going as close to 0.0 without being exactly 0.0
        // if you have a better idea. Please be my guest
        return 0.0000001 / (180 / pi);
      case JoystickMoveDirectional.MOVE_UP:
        return -90 / (180 / pi);
      case JoystickMoveDirectional.MOVE_DOWN:
        return 90 / (180 / pi);
      case JoystickMoveDirectional.MOVE_UP_LEFT:
        return -135 / (180 / pi);
      case JoystickMoveDirectional.MOVE_UP_RIGHT:
        return -45 / (180 / pi);
      case JoystickMoveDirectional.MOVE_DOWN_LEFT:
        return 135 / (180 / pi);
      case JoystickMoveDirectional.MOVE_DOWN_RIGHT:
        return 45 / (180 / pi);
      default:
        return 0;
    }
  }

  @override
  void idle() {
    if (this is JoystickListener) {
      (this as JoystickListener).currentDirectional =
          JoystickMoveDirectional.IDLE;
    }
    super.idle();
  }
}
